@@ -18,16 +18,41 @@ from collections import abc
 import contextlib
 import threading
 
+from tensorflow.python.distribute import distribute_lib
 from tensorflow.python.distribute import tpu_values as tpu_values_lib
 from tensorflow.python.distribute import values as values_lib
+from tensorflow.python.distribute.reduce_util import ReduceOp
 from tensorflow.python.eager import context
-from tensorflow.python.eager import tape
+from tensorflow.python.eager import record
+from tensorflow.python.framework import composite_tensor
 from tensorflow.python.framework import ops
 from tensorflow.python.framework import tensor_util
 from tensorflow.python.ops import array_ops
 from tensorflow.python.ops import control_flow_ops
+from tensorflow.python.ops import resource_variable_ops
 from tensorflow.python.ops import variable_scope as vs
+from tensorflow.python.ops.losses import losses_impl
 from tensorflow.python.util import nest
+from tensorflow.python.util.tf_export import tf_export
+
+
+@tf_export(v1=["distribute.get_loss_reduction"])
+def get_loss_reduction():
+  """`tf.distribute.ReduceOp` corresponding to the last loss reduction.
+
+  Returns:
+    `tf.distribute.ReduceOp` corresponding to the last loss reduction for
+    estimator and v1 optimizer use case. `tf.distribute.ReduceOp.SUM` otherwise.
+  """
+  if not distribute_lib.get_strategy()._scale_loss_for_estimator:  # pylint: disable=protected-access
+    # If we are not in Estimator context then return 'SUM'. We do not need to
+    # scale loss in the optimizer.
+    return ReduceOp.SUM
+  last_reduction = ops.get_default_graph()._last_loss_reduction  # pylint: disable=protected-access
+  if (last_reduction == losses_impl.Reduction.SUM or
+      last_reduction == "sum"):  # Check for tf.keras.losses.Reduction.SUM
+    return ReduceOp.SUM
+  return ReduceOp.MEAN
 
 
 def regroup(values, wrap_class=values_lib.PerReplica, always_wrap=False):
@@ -98,13 +123,13 @@ def regroup(values, wrap_class=values_lib.PerReplica, always_wrap=False):
   if same_id and isinstance(v0, values_lib.DistributedVariable):
     return v0
   # * If v0 is a member of a distributed variable, in which case
-  #   hasattr(v0, "_distributed_container") is true, we want to
+  #   value_container(v0) is not v0 itself, we want to
   #   return the DistributedVariable that contains it using the
   #   _distributed_container logic below. This case can trigger
   #   same_id when there is only one device.
   # * In any other situation, same_id means we return v0 unless `always_wrap` is
   #   true.
-  if same_id and not always_wrap and not hasattr(v0, "_distributed_container"):
+  if same_id and not always_wrap and value_container(v0) is v0:
     return v0
 
   # Detect the case where each device has a parallel component of the
@@ -112,15 +137,17 @@ def regroup(values, wrap_class=values_lib.PerReplica, always_wrap=False):
   # want to return the containing MirroredVariable, after a bunch of
   # sanity checking. In particular, each component should have the
   # same container, and the devices of the variables should match the
-  # keys of the per-replica dictionary.
-  if hasattr(v0, "_distributed_container"):
+  # keys of the per-replica dictionary. For _UnreadVariables, use the wrap_class
+  # path, which calls tf.identity on them.
+  if (not isinstance(v0, resource_variable_ops._UnreadVariable) and  # pylint: disable=protected-access
+      value_container(v0) is not v0):
     # pylint: disable=protected-access
     assert not isinstance(v0, values_lib.MirroredVariable), (
         "ids = %s, values = %s" % ([id(v) for v in values], values))
-    distributed_container = v0._distributed_container()
+    distributed_container = value_container(v0)
     assert distributed_container is not None
     for v in values[1:]:
-      assert distributed_container is v._distributed_container()
+      assert distributed_container is value_container(v)
     return distributed_container
   # pylint: enable=protected-access
 
@@ -206,14 +233,19 @@ def value_container(val):
     If value does not belong to any container (including the case of
     container having been destroyed), returns the value itself.
   """
-  if (hasattr(val, "_distributed_container") and
-      # DistributedVariable has _distributed_container defined
-      # but we don't want to return it.
-      not isinstance(val, values_lib.DistributedVariable)):
-    container = val._distributed_container()  # pylint: disable=protected-access
-    if container is not None:
-      return container
-  return val
+  # DistributedVariable has _distributed_container defined but we don't want to
+  # return it.
+  container = None
+  if not isinstance(val, values_lib.DistributedVariable):
+    if hasattr(val, "_distributed_container"):
+      container = val._distributed_container()  # pylint: disable=protected-access
+    elif (isinstance(val, composite_tensor.CompositeTensor) and
+          hasattr(val, "handle") and
+          hasattr(val.handle, "_distributed_container")):
+      # For ResourceVariables, the _distributed_container attribute
+      # is added to their handle tensors.
+      container = val.handle._distributed_container()  # pylint: disable=protected-access
+  return container if container is not None else val
 
 
 def is_distributed_variable(v):
@@ -228,7 +260,7 @@ def is_distributed_table(v):
 
 def _validate_colocate_extended(v, extended):
   variable_strategy = v._distribute_strategy  # pylint: disable=protected-access
-  if variable_strategy.extended is not extended:
+  if not variable_strategy or variable_strategy.extended is not extended:
     raise ValueError(
         "`colocate_vars_with` must only be passed a variable created in this "
         "tf.distribute.Strategy.scope(), not %s created in scope: %s" %
@@ -289,6 +321,12 @@ def create_mirrored_variable(strategy, real_mirrored_creator, class_mapping,
   """Create distributed variables with given synchronization and aggregation."""
   # Figure out what collections this variable should be added to.
   # We'll add the MirroredVariable to those collections instead.
+
+  if kwargs.pop("experimental_batch_initialization", None):
+    variable_class_key = "LazyVariableClass"
+  else:
+    variable_class_key = "VariableClass"
+
   var_collections = kwargs.pop("collections", None)
   if var_collections is None:
     var_collections = [ops.GraphKeys.GLOBAL_VARIABLES]
@@ -307,7 +345,7 @@ def create_mirrored_variable(strategy, real_mirrored_creator, class_mapping,
   # TODO(josh11b,apassos): It would be better if variable initialization
   # was never recorded on the tape instead of having to do this manually
   # here.
-  with tape.stop_recording():
+  with record.stop_recording():
     value_list = real_mirrored_creator(**kwargs)
     # MirroredVariable is recreated during saved_model loading, and its
     # component variables (value_list) will have None initializer. We
@@ -322,7 +360,7 @@ def create_mirrored_variable(strategy, real_mirrored_creator, class_mapping,
     if use_var_policy:
       var_policy_cls = policy_mapping.get(synchronization)
       var_policy = var_policy_cls(aggregation=aggregation)
-      var_cls = class_mapping.get("VariableClass")
+      var_cls = class_mapping.get(variable_class_key)
       result = var_cls(strategy, value_list, aggregation, var_policy=var_policy)
     else:
       var_cls = class_mapping.get(synchronization)
@@ -358,17 +396,11 @@ def create_mirrored_variable(strategy, real_mirrored_creator, class_mapping,
 # Return True if the Value is Mirrored or the Variable is replicated and kept in
 # sync.
 def is_mirrored(val):
-  if isinstance(val, values_lib.DistributedVariable):
-    if val._policy:  # pylint: disable=protected-access
-      return val._policy._is_mirrored()  # pylint: disable=protected-access
-  return isinstance(val, values_lib.Mirrored)
+  return (getattr(val, "_is_mirrored", lambda: False))()
 
 
 def is_sync_on_read(val):
-  if isinstance(val, values_lib.DistributedVariable):
-    if val._policy:  # pylint: disable=protected-access
-      return not val._policy._is_mirrored()  # pylint: disable=protected-access
-  return not isinstance(val, values_lib.Mirrored)
+  return not is_mirrored(val)
 
 
 class CachingScopeLocal(threading.local):
@@ -458,6 +490,7 @@ TPU_VARIABLE_POLICY_MAPPING = {
 
 TPU_VARIABLE_CLASS_MAPPING = {
     "VariableClass": tpu_values_lib.TPUDistributedVariable,
+    "LazyVariableClass": tpu_values_lib.TPULazyDistributedVariable,
     vs.VariableSynchronization.ON_WRITE: tpu_values_lib.TPUMirroredVariable,
     vs.VariableSynchronization.ON_READ: tpu_values_lib.TPUSyncOnReadVariable,
 }
