@@ -17,17 +17,22 @@
 import abc
 import math
 import typing
-from typing import Any, Dict, Callable, List, Optional, Text, Tuple, TypeVar, Union
+from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Text, Tuple, TypeVar, Union
 
 from absl import logging
-import six
 
 from tensorflow.core.protobuf.tpu import optimization_parameters_pb2
 from tensorflow.core.protobuf.tpu import tpu_embedding_configuration_pb2
+from tensorflow.python.distribute import device_util
 from tensorflow.python.distribute import sharded_variable
+from tensorflow.python.distribute import tpu_strategy
+from tensorflow.python.framework import device_spec
+from tensorflow.python.framework import dtypes
 from tensorflow.python.framework import ops
 from tensorflow.python.framework.tensor_shape import TensorShape
+from tensorflow.python.ops import array_ops
 from tensorflow.python.ops import init_ops_v2
+from tensorflow.python.ops import math_ops
 from tensorflow.python.ops import variables as tf_variables
 from tensorflow.python.tpu.ops import tpu_ops
 from tensorflow.python.types import core
@@ -42,8 +47,7 @@ SlotVarCreationFnType = Callable[
 ClipValueType = Union[Tuple[float, float], float]
 
 
-@six.add_metaclass(abc.ABCMeta)
-class _Optimizer(object):
+class _Optimizer(metaclass=abc.ABCMeta):
   """Base class for all optimizers, with common parameters."""
 
   def __init__(
@@ -55,7 +59,9 @@ class _Optimizer(object):
       weight_decay_factor: Optional[float],
       multiply_weight_decay_factor_by_learning_rate: bool,
       clipvalue: Optional[ClipValueType] = None,
-      slot_variable_creation_fn: Optional[SlotVarCreationFnType] = None):
+      slot_variable_creation_fn: Optional[SlotVarCreationFnType] = None,
+      low_dimensional_packing_status: bool = False,
+  ):
     self.learning_rate = learning_rate
     self.use_gradient_accumulation = use_gradient_accumulation
     self.clip_weight_min = clip_weight_min
@@ -81,6 +87,7 @@ class _Optimizer(object):
           f"Argument `slot_variable_creation_fn` must be either None or a "
           f"callable. Received: {slot_variable_creation_fn}")
     self.slot_variable_creation_fn = slot_variable_creation_fn
+    self.low_dimensional_packing_status = low_dimensional_packing_status
 
   @abc.abstractmethod
   def _slot_names(self) -> List[Text]:
@@ -127,6 +134,10 @@ class _Optimizer(object):
       if self.multiply_weight_decay_factor_by_learning_rate:
         parameters.multiply_weight_decay_factor_by_learning_rate = True
 
+    parameters.low_dimensional_packing_status = (
+        self.low_dimensional_packing_status
+    )
+
   @abc.abstractmethod
   def _load(self) -> Callable[..., ops.Operation]:
     """Returns the load function for the optimizer."""
@@ -175,6 +186,167 @@ class _Optimizer(object):
     return hash(tuple(self.__dict__.items()))
 
 
+@tf_export("tpu.experimental.embedding.CustomOptimizer")
+class CustomOptimizer(_Optimizer):
+  """Optimization parameters for custom optimizer for TPU embeddings.
+
+  This optimizer gives the user the ability to define a custom optimizer
+  for running embedding lookups on TPU v5p.
+
+  The custom computation should be a function which takes gradient, embedding
+  table, a list of slot variables, learning_rate and a list of hyperparameters.
+  The function should perform the gradient update on the embedding_table +
+  slot_variables and return the updated embedding_table and slot_variables. e.g.
+
+  ```python
+  @tf.function
+  def sgd_optimizer_computation(
+      gradient,
+      embedding_table,
+      slot_variables,
+      learning_rate,
+      hyperparameters,
+  ):
+    del slot_variables, hyperparameters
+    return embedding_table - gradient * learning_rate
+  ```
+
+  Above is a simple example of a sgd optimizer. You can also define a more
+  complex optimizer which updates multiple tables and slot variables.
+
+  ```python
+  @def_function.function
+  def adagrad_optimizer_computation(
+      gradient,
+      embedding_table,
+      slot_variables,
+      learning_rate,
+      hyperparameters,
+  ):
+    del hyperparameters
+    accumulator = slot_variables[0]
+    new_accumulator = accumulator + gradient * gradient
+    updated_embedding_table = (
+        embedding_table
+        - learning_rate * gradient / math_ops.sqrt(new_accumulator)
+    )
+    return (updated_embedding_table, new_accumulator)
+  ```
+
+  The custom computation is defined as a per-row update function and it will be
+  auto scaled for the entire table (slot variables).
+
+  NOTE: This optimizer can only be used with the `TPUEmbeddingV2` class.
+
+  Pass this to `tf.tpu.experimental.embedding.TPUEmbeddingV2` via the
+  `optimizer` argument to set the global optimizer and its parameters:
+
+  ```python
+  optimizer = tf.tpu.experimental.embedding.CustomOptimizer(
+        custom_computation=adagrad_optimizer_computation,
+        learning_rate=1.0,
+        slot_names=['accumulators'],
+        slot_initializers=[
+            tf.constant_initializer(0.1, support_partition=True)
+        ],
+    )
+  embedding = tf.tpu.experimental.embedding.TPUEmbeddingV2(
+      ...
+      optimizer=optimizer)
+  ```
+
+  This can also be used in a `tf.tpu.experimental.embedding.TableConfig` as the
+  optimizer parameter to set a table specific optimizer. This will override the
+  optimizer and parameters for global embedding optimizer defined above:
+
+  ```python
+  table_one = tf.tpu.experimental.embedding.TableConfig(
+      vocabulary_size=...,
+      dim=...,
+      optimizer=optimizer)
+  table_two = tf.tpu.experimental.embedding.TableConfig(
+      vocabulary_size=...,
+      dim=...)
+
+  feature_config = (
+      tf.tpu.experimental.embedding.FeatureConfig(
+          table=table_one),
+      tf.tpu.experimental.embedding.FeatureConfig(
+          table=table_two))
+
+  embedding = tf.tpu.experimental.embedding.TPUEmbedding(
+      feature_config=feature_config,
+      batch_size=...
+      optimizer=optimizer)
+  ```
+  In this example, the optimizer of the first table will be the one specified
+  in the table config. The second table will use the optimizer specified in the
+  TPUEmbedding argument.
+  """
+
+  def __init__(
+      self,
+      custom_computation: core.PolymorphicFunction,
+      learning_rate: Union[float, Callable[[], float]] = 0.01,
+      slot_names: Optional[List[str]] = None,
+      slot_initializers: Optional[List[init_ops_v2.Initializer]] = None,
+      hyperparameters: Optional[List[Union[float, Callable[[], float]]]] = None,
+  ) -> Any:
+    super().__init__(  # pytype: disable=wrong-arg-types
+        learning_rate,
+        use_gradient_accumulation=False,
+        clip_weight_min=None,
+        clip_weight_max=None,
+        weight_decay_factor=None,
+        multiply_weight_decay_factor_by_learning_rate=None,
+        clipvalue=None,
+        slot_variable_creation_fn=None,
+        low_dimensional_packing_status=False,
+    )
+    # We need to convert the slot names and initializers to tuples to make
+    # them hashable.
+    self._slot_names_attr = tuple(slot_names if slot_names else ())
+    self._slot_initializers_attr = tuple(
+        slot_initializers if slot_initializers else ()
+    )
+    num_slot_names = len(self._slot_names_attr)
+    num_slot_initializers = len(self._slot_initializers_attr)
+    if num_slot_names != num_slot_initializers:
+      raise ValueError(
+          f"The number of slot_names ({num_slot_names}) must match"
+          " the number of slot_initializers"
+          f" ({num_slot_initializers})."
+      )
+    self._hyperparameters = tuple(hyperparameters if hyperparameters else ())
+    self._custom_computation = custom_computation
+
+  def _slot_names(self) -> List[Text]:
+    return list(self._slot_names_attr)
+
+  def _slot_initializers(self) -> List[init_ops_v2.Initializer]:
+    return list(self._slot_initializers_attr)
+
+  def _load(self) -> Callable[..., ops.Operation]:
+    raise NotImplementedError(
+        "Custom optimizer does not support load op since it is only used for"
+        " TPUEmbeddingV2."
+    )
+
+  def _retrieve(self) -> Callable[..., core.Tensor]:
+    raise NotImplementedError(
+        "Custom optimizer does not support retrieve op since it is only used"
+        " for TPUEmbeddingV2."
+    )
+
+  @property
+  def hyperparameters(self) -> Tuple[Union[float, Callable[[], float]], ...]:
+    return self._hyperparameters
+
+  @property
+  def custom_computation(self) -> core.ConcreteFunction:
+    return self._custom_computation
+
+
 @tf_export("tpu.experimental.embedding.SGD")
 class SGD(_Optimizer):
   """Optimization parameters for stochastic gradient descent for TPU embeddings.
@@ -222,14 +394,17 @@ class SGD(_Optimizer):
   algorithm.
   """
 
-  def __init__(self,
-               learning_rate: Union[float, Callable[[], float]] = 0.01,
-               use_gradient_accumulation: bool = True,
-               clip_weight_min: Optional[float] = None,
-               clip_weight_max: Optional[float] = None,
-               weight_decay_factor: Optional[float] = None,
-               multiply_weight_decay_factor_by_learning_rate: bool = None,
-               clipvalue: Optional[ClipValueType] = None):
+  def __init__(
+      self,
+      learning_rate: Union[float, Callable[[], float]] = 0.01,
+      use_gradient_accumulation: bool = True,
+      clip_weight_min: Optional[float] = None,
+      clip_weight_max: Optional[float] = None,
+      weight_decay_factor: Optional[float] = None,
+      multiply_weight_decay_factor_by_learning_rate: Optional[bool] = None,
+      clipvalue: Optional[ClipValueType] = None,
+      low_dimensional_packing_status: bool = False,
+  ):
     """Optimization parameters for stochastic gradient descent.
 
     Args:
@@ -253,11 +428,22 @@ class SGD(_Optimizer):
         accuracy). See
         'tensorflow/core/protobuf/tpu/optimization_parameters.proto' for more
         information on gradient accumulation and its impact on tpu embeddings.
+      low_dimensional_packing_status: Status of the low-dimensional embedding
+        packing optimization controls whether to optimize the packing of
+        1-dimensional, 2-dimensional, and 4-dimensional embedding tables in
+        memory.
     """
-    super(SGD, self).__init__(
-        learning_rate, use_gradient_accumulation, clip_weight_min,
-        clip_weight_max, weight_decay_factor,
-        multiply_weight_decay_factor_by_learning_rate, clipvalue)
+    super().__init__(
+        learning_rate,
+        use_gradient_accumulation,
+        clip_weight_min,
+        clip_weight_max,
+        weight_decay_factor,
+        multiply_weight_decay_factor_by_learning_rate,
+        clipvalue,
+        None,
+        low_dimensional_packing_status,
+    )
 
   def _slot_names(self) -> List[Text]:
     return []
@@ -267,7 +453,7 @@ class SGD(_Optimizer):
 
   def _set_optimization_parameters(
       self, parameters: optimization_parameters_pb2.OptimizationParameters):
-    super(SGD, self)._set_optimization_parameters(parameters)
+    super()._set_optimization_parameters(parameters)
     parameters.stochastic_gradient_descent.SetInParent()
 
   def _load(self) -> Callable[..., ops.Operation]:
@@ -332,9 +518,11 @@ class Adagrad(_Optimizer):
       clip_weight_min: Optional[float] = None,
       clip_weight_max: Optional[float] = None,
       weight_decay_factor: Optional[float] = None,
-      multiply_weight_decay_factor_by_learning_rate: bool = None,
+      multiply_weight_decay_factor_by_learning_rate: Optional[bool] = None,
       slot_variable_creation_fn: Optional[SlotVarCreationFnType] = None,
-      clipvalue: Optional[ClipValueType] = None):
+      clipvalue: Optional[ClipValueType] = None,
+      low_dimensional_packing_status: bool = False,
+  ):
     """Optimization parameters for Adagrad.
 
     Args:
@@ -351,21 +539,31 @@ class Adagrad(_Optimizer):
         `weight_decay_factor` is multiplied by the current learning rate.
       slot_variable_creation_fn: If you wish do directly control the creation of
         the slot variables, set this to a callable taking three parameters: a
-          table variable, a list of slot names to create for it, and a list of
-          initializers. This function should return a dict with the slot names
-          as keys and the created variables as values with types matching the
-          table variable. When set to None (the default), uses the built-in
-          variable creation.
+        table variable, a list of slot names to create for it, and a list of
+        initializers. This function should return a dict with the slot names as
+        keys and the created variables as values with types matching the table
+        variable. When set to None (the default), uses the built-in variable
+        creation.
       clipvalue: Controls clipping of the gradient. Set to either a single
         positive scalar value to get clipping or a tuple of scalar values (min,
         max) to set a separate maximum or minimum. If one of the two entries is
         None, then there will be no clipping that direction.
+      low_dimensional_packing_status: Status of the low-dimensional embedding
+        packing optimization controls whether to optimize the packing of
+        1-dimensional, 2-dimensional, and 4-dimensional embedding tables in
+        memory.
     """
-    super(Adagrad, self).__init__(
-        learning_rate, use_gradient_accumulation, clip_weight_min,
-        clip_weight_max, weight_decay_factor,
-        multiply_weight_decay_factor_by_learning_rate, clipvalue,
-        slot_variable_creation_fn)
+    super().__init__(
+        learning_rate,
+        use_gradient_accumulation,
+        clip_weight_min,
+        clip_weight_max,
+        weight_decay_factor,
+        multiply_weight_decay_factor_by_learning_rate,
+        clipvalue,
+        slot_variable_creation_fn,
+        low_dimensional_packing_status,
+    )
     if initial_accumulator_value <= 0:
       raise ValueError(
           f"Argument `initial_accumulator_value` must be a positive float. "
@@ -376,11 +574,15 @@ class Adagrad(_Optimizer):
     return ["accumulators"]
 
   def _slot_initializers(self) -> List[init_ops_v2.Initializer]:
-    return [init_ops_v2.Constant(self.initial_accumulator_value)]
+    return [
+        init_ops_v2.Constant(
+            self.initial_accumulator_value, support_partition=True
+        )
+    ]
 
   def _set_optimization_parameters(
       self, parameters: optimization_parameters_pb2.OptimizationParameters):
-    super(Adagrad, self)._set_optimization_parameters(parameters)
+    super()._set_optimization_parameters(parameters)
     parameters.adagrad.SetInParent()
 
   def _load(self) -> Callable[..., ops.Operation]:
@@ -449,9 +651,11 @@ class AdagradMomentum(_Optimizer):
       clip_weight_min: Optional[float] = None,
       clip_weight_max: Optional[float] = None,
       weight_decay_factor: Optional[float] = None,
-      multiply_weight_decay_factor_by_learning_rate: bool = None,
+      multiply_weight_decay_factor_by_learning_rate: Optional[bool] = None,
       slot_variable_creation_fn: Optional[SlotVarCreationFnType] = None,
-      clipvalue: Optional[ClipValueType] = None):
+      clipvalue: Optional[ClipValueType] = None,
+      low_dimensional_packing_status: bool = False,
+  ):
     """Optimization parameters for Adagrad + Momentum.
 
     Args:
@@ -473,21 +677,31 @@ class AdagradMomentum(_Optimizer):
         `weight_decay_factor` is multiplied by the current learning rate.
       slot_variable_creation_fn: If you wish do directly control the creation of
         the slot variables, set this to a callable taking three parameters: a
-          table variable, a list of slot names to create for it, and a list of
-          initializers. This function should return a dict with the slot names
-          as keys and the created variables as values with types matching the
-          table variable. When set to None (the default), uses the built-in
-          variable creation.
+        table variable, a list of slot names to create for it, and a list of
+        initializers. This function should return a dict with the slot names as
+        keys and the created variables as values with types matching the table
+        variable. When set to None (the default), uses the built-in variable
+        creation.
       clipvalue: Controls clipping of the gradient. Set to either a single
         positive scalar value to get clipping or a tuple of scalar values (min,
         max) to set a separate maximum or minimum. If one of the two entries is
         None, then there will be no clipping that direction.
+      low_dimensional_packing_status: Status of the low-dimensional embedding
+        packing optimization controls whether to optimize the packing of
+        1-dimensional, 2-dimensional, and 4-dimensional embedding tables in
+        memory.
     """
-    super(AdagradMomentum,
-          self).__init__(learning_rate, use_gradient_accumulation,
-                         clip_weight_min, clip_weight_max, weight_decay_factor,
-                         multiply_weight_decay_factor_by_learning_rate,
-                         clipvalue, slot_variable_creation_fn)
+    super().__init__(
+        learning_rate,
+        use_gradient_accumulation,
+        clip_weight_min,
+        clip_weight_max,
+        weight_decay_factor,
+        multiply_weight_decay_factor_by_learning_rate,
+        clipvalue,
+        slot_variable_creation_fn,
+        low_dimensional_packing_status,
+    )
     if epsilon <= 0:
       raise ValueError("Adagrad momentum: epsilon must be positive")
     if exponent <= 0:
@@ -502,11 +716,15 @@ class AdagradMomentum(_Optimizer):
     return ["accumulators", "momenta"]
 
   def _slot_initializers(self) -> List[init_ops_v2.Initializer]:
-    return [init_ops_v2.Constant(), init_ops_v2.Constant()]
+    return [
+        init_ops_v2.Constant(support_partition=True),
+        init_ops_v2.Constant(support_partition=True),
+    ]
 
   def _set_optimization_parameters(
-      self, parameters: optimization_parameters_pb2.OptimizationParameters):
-    super(AdagradMomentum, self)._set_optimization_parameters(parameters)
+      self, parameters: optimization_parameters_pb2.OptimizationParameters
+  ):
+    super()._set_optimization_parameters(parameters)
     parameters.adagrad_momentum.SetInParent()
     parameters.adagrad_momentum.momentum = self.momentum
     parameters.adagrad_momentum.use_nesterov = self.use_nesterov
@@ -583,11 +801,13 @@ class FTRL(_Optimizer):
       clip_weight_min: Optional[float] = None,
       clip_weight_max: Optional[float] = None,
       weight_decay_factor: Optional[float] = None,
-      multiply_weight_decay_factor_by_learning_rate: bool = None,
+      multiply_weight_decay_factor_by_learning_rate: Optional[bool] = None,
       slot_variable_creation_fn: Optional[SlotVarCreationFnType] = None,
       clipvalue: Optional[ClipValueType] = None,
       multiply_linear_by_learning_rate: bool = False,
-      allow_zero_accumulator: bool = False):
+      allow_zero_accumulator: bool = False,
+      low_dimensional_packing_status: bool = False,
+  ):
     """Optimization parameters for Adagrad.
 
     Args:
@@ -613,11 +833,11 @@ class FTRL(_Optimizer):
         `weight_decay_factor` is multiplied by the current learning rate.
       slot_variable_creation_fn: If you wish do directly control the creation of
         the slot variables, set this to a callable taking three parameters: a
-          table variable, a list of slot names to create for it, and a list of
-          initializers. This function should return a dict with the slot names
-          as keys and the created variables as values with types matching the
-          table variable. When set to None (the default), uses the built-in
-          variable creation.
+        table variable, a list of slot names to create for it, and a list of
+        initializers. This function should return a dict with the slot names as
+        keys and the created variables as values with types matching the table
+        variable. When set to None (the default), uses the built-in variable
+        creation.
       clipvalue: Controls clipping of the gradient. Set to either a single
         positive scalar value to get clipping or a tuple of scalar values (min,
         max) to set a separate maximum or minimum. If one of the two entries is
@@ -635,11 +855,22 @@ class FTRL(_Optimizer):
         allow zero and near-zero accumulator values at the cost of some
         performance; this only needs to be set if you are using an initial
         accumulator value of zero, which is uncommon.
+      low_dimensional_packing_status: Status of the low-dimensional embedding
+        packing optimization controls whether to optimize the packing of
+        1-dimensional, 2-dimensional, and 4-dimensional embedding tables in
+        memory.
     """
-    super().__init__(learning_rate, use_gradient_accumulation, clip_weight_min,
-                     clip_weight_max, weight_decay_factor,
-                     multiply_weight_decay_factor_by_learning_rate, clipvalue,
-                     slot_variable_creation_fn)
+    super().__init__(
+        learning_rate,
+        use_gradient_accumulation,
+        clip_weight_min,
+        clip_weight_max,
+        weight_decay_factor,
+        multiply_weight_decay_factor_by_learning_rate,
+        clipvalue,
+        slot_variable_creation_fn,
+        low_dimensional_packing_status,
+    )
     if initial_accumulator_value <= 0:
       raise ValueError(
           f"Argument `initial_accumulator_value` must be a positive float. "
@@ -657,12 +888,15 @@ class FTRL(_Optimizer):
 
   def _slot_initializers(self) -> List[init_ops_v2.Initializer]:
     return [
-        init_ops_v2.Constant(self.initial_accumulator_value),
-        init_ops_v2.Constant()
+        init_ops_v2.Constant(
+            self.initial_accumulator_value, support_partition=True
+        ),
+        init_ops_v2.Constant(support_partition=True),
     ]
 
   def _set_optimization_parameters(
-      self, parameters: optimization_parameters_pb2.OptimizationParameters):
+      self, parameters: optimization_parameters_pb2.OptimizationParameters
+  ):
     super()._set_optimization_parameters(parameters)
     ftrl = parameters.ftrl
     ftrl.l1 = self.l1_regularization_strength
@@ -742,9 +976,11 @@ class Adam(_Optimizer):
       clip_weight_min: Optional[float] = None,
       clip_weight_max: Optional[float] = None,
       weight_decay_factor: Optional[float] = None,
-      multiply_weight_decay_factor_by_learning_rate: bool = None,
+      multiply_weight_decay_factor_by_learning_rate: Optional[bool] = None,
       slot_variable_creation_fn: Optional[SlotVarCreationFnType] = None,
-      clipvalue: Optional[ClipValueType] = None):
+      clipvalue: Optional[ClipValueType] = None,
+      low_dimensional_packing_status: bool = False,
+  ):
     """Optimization parameters for Adam.
 
     See 'tensorflow/core/protobuf/tpu/optimization_parameters.proto' for a
@@ -774,21 +1010,31 @@ class Adam(_Optimizer):
         `weight_decay_factor` is multiplied by the current learning rate.
       slot_variable_creation_fn: If you wish do directly control the creation of
         the slot variables, set this to a callable taking three parameters: a
-          table variable, a list of slot names to create for it, and a list of
-          initializers. This function should return a dict with the slot names
-          as keys and the created variables as values with types matching the
-          table variable. When set to None (the default), uses the built-in
-          variable creation.
+        table variable, a list of slot names to create for it, and a list of
+        initializers. This function should return a dict with the slot names as
+        keys and the created variables as values with types matching the table
+        variable. When set to None (the default), uses the built-in variable
+        creation.
       clipvalue: Controls clipping of the gradient. Set to either a single
         positive scalar value to get clipping or a tiple of scalar values (min,
         max) to set a separate maximum or minimum. If one of the two entries is
         None, then there will be no clipping that direction.
+      low_dimensional_packing_status: Status of the low-dimensional embedding
+        packing optimization controls whether to optimize the packing of
+        1-dimensional, 2-dimensional, and 4-dimensional embedding tables in
+        memory.
     """
     super(Adam, self).__init__(
-        learning_rate, use_gradient_accumulation, clip_weight_min,
-        clip_weight_max, weight_decay_factor,
-        multiply_weight_decay_factor_by_learning_rate, clipvalue,
-        slot_variable_creation_fn)
+        learning_rate,
+        use_gradient_accumulation,
+        clip_weight_min,
+        clip_weight_max,
+        weight_decay_factor,
+        multiply_weight_decay_factor_by_learning_rate,
+        clipvalue,
+        slot_variable_creation_fn,
+        low_dimensional_packing_status,
+    )
     if beta_1 < 0. or beta_1 >= 1.:
       raise ValueError(
           f"Argument `beta_1` must be >= 0 and < 1. Received: {beta_1}.")
@@ -813,10 +1059,14 @@ class Adam(_Optimizer):
     return ["momenta", "velocities"]
 
   def _slot_initializers(self) -> List[init_ops_v2.Initializer]:
-    return [init_ops_v2.Constant(), init_ops_v2.Constant()]
+    return [
+        init_ops_v2.Constant(support_partition=True),
+        init_ops_v2.Constant(support_partition=True),
+    ]
 
   def _set_optimization_parameters(
-      self, parameters: optimization_parameters_pb2.OptimizationParameters):
+      self, parameters: optimization_parameters_pb2.OptimizationParameters
+  ):
     super(Adam, self)._set_optimization_parameters(parameters)
     parameters.adam.beta1 = self.beta_1
     parameters.adam.beta2 = self.beta_2
@@ -831,8 +1081,72 @@ class Adam(_Optimizer):
     return tpu_ops.retrieve_tpu_embedding_adam_parameters
 
 
+@tf_export("tpu.experimental.embedding.QuantizationConfig")
+class QuantizationConfig:
+  """Settings for simulated quantization of the tpu embedding table.
+
+  When simulated quantization is enabled, the results of the embedding lookup
+  are clipped and quantized according to the settings here before the combiner
+  is applied.
+
+  For example, to quantize `input` the following is done:
+  ```python
+  if input < lower
+    input = lower
+  if input > upper
+    input = upper
+  quantum = (upper - lower) / (num_buckets - 1)
+  input = math.floor((input - lower) / quantum + 0.5) * quantum + lower
+  ```
+
+  See tensorflow/core/protobuf/tpu/optimization_parameters.proto for more
+  details.
+
+  NOTE: This does not change the storage type of the embedding table, that will
+  continue to be float32 as will the saved variable in the checkpoint. You will
+  have to manually quantize the variable (typically with the same algorithm and
+  settings as above) manually.
+  """
+
+  def __init__(self, num_buckets: int, lower: float, upper: float):
+    """Simulated quantizaiton configuration.
+
+    Args:
+      num_buckets: The number of quantization buckets, must be atleast 2.
+      lower: The lower bound for the quantization range.
+      upper: The upper bound for the quantization range.
+
+    Returns:
+      `QuantizationConfig`.
+
+    Raises:
+      ValueError: if `num_buckets` is less than 2.
+    """
+    if num_buckets < 2:
+      raise ValueError(f"num_buckets is {num_buckets}, must be at least 2 for "
+                       f"simulated quantization.")
+
+    self.num_buckets = num_buckets
+    self.lower = lower
+    self.upper = upper
+
+  def _set_optimization_parameters(
+      self, parameters: optimization_parameters_pb2.OptimizationParameters):
+    parameters.simulated_quantization.enabled = True
+    parameters.simulated_quantization.num_buckets = self.num_buckets
+    parameters.simulated_quantization.clipping_limits.lower.value = self.lower
+    parameters.simulated_quantization.clipping_limits.upper.value = self.upper
+
+  def __repr__(self):
+    return ("QuantizationConfig(num_buckets={num_buckets!r}, lower={lower!r}, "
+            "upper={upper!r})".format(
+                num_buckets=self.num_buckets,
+                lower=self.lower,
+                upper=self.upper))
+
+
 @tf_export("tpu.experimental.embedding.TableConfig")
-class TableConfig(object):
+class TableConfig:
   """Configuration data for one embedding table.
 
   This class holds the configuration data for a single embedding table. It is
@@ -868,13 +1182,19 @@ class TableConfig(object):
 
   """
 
-  def __init__(self,
-               vocabulary_size: int,
-               dim: int,
-               initializer: Optional[Callable[[Any], None]] = None,
-               optimizer: Optional[_Optimizer] = None,
-               combiner: Text = "mean",
-               name: Optional[Text] = None):
+  def __init__(
+      self,
+      vocabulary_size: int,
+      dim: int,
+      initializer: Optional[Callable[[Any], None]] = None,
+      optimizer: Optional[_Optimizer] = None,
+      combiner: Text = "mean",
+      name: Optional[Text] = None,
+      quantization_config: QuantizationConfig = None,
+      # TODO(b/295372790): Change the type to SparseCoreTableLayout after it is
+      # open sourced.
+      layout: Optional[Any] = None,
+  ):
     """Embedding table configuration.
 
     Args:
@@ -888,14 +1208,21 @@ class TableConfig(object):
       optimizer: An optional instance of an optimizer parameters class, instance
         of one of `tf.tpu.experimental.embedding.SGD`,
         `tf.tpu.experimental.embedding.Adagrad` or
-        `tf.tpu.experimental.embedding.Adam`. It set will override the global
+        `tf.tpu.experimental.embedding.Adam`. If set will override the global
         optimizer passed to `tf.tpu.experimental.embedding.TPUEmbedding`.
       combiner: A string specifying how to reduce if there are multiple entries
         in a single row. Currently 'mean', 'sqrtn', 'sum' are supported, with
         'mean' the default. 'sqrtn' often achieves good accuracy, in particular
         with bag-of-words columns. For more information, see
         `tf.nn.embedding_lookup_sparse`.
-      name: An optional string used to name the table. Useful for debugging.
+      name: An optional string used to name the table. Must be defined if
+        running on SparseCore.
+      quantization_config: The simulated quantization config. An instance of
+        `tf.tpu.experimental.embedding.QuantizationConfig`. See the class for
+        more documentation.
+      layout: If the table already has its layout computed, you can pass it in
+        here. Otherwise, we will compute it for you. Most users should leave
+        this as None.
 
     Returns:
       `TableConfig`.
@@ -929,12 +1256,20 @@ class TableConfig(object):
           f"Argument `combiner` must be one of {accepted_combiners}. "
           f"Received: {combiner}")
 
+    if name is None:
+      logging.warning(
+          "Name of the table config must be specified for running on"
+          " SparseCore. Different table configs must have unique names."
+      )
+
     self.vocabulary_size = vocabulary_size
     self.dim = dim
     self.initializer = initializer
     self.optimizer = optimizer
     self.combiner = combiner
     self.name = name
+    self.quantization_config = quantization_config
+    self.layout = layout
 
   def __repr__(self):
     # If using the default initializer, just print "None" for clarity.
@@ -944,24 +1279,75 @@ class TableConfig(object):
       # PY2 type checking can't infer type of initializer even after if.
       initializer = typing.cast(init_ops_v2.TruncatedNormal, initializer)
       if (initializer.mean == 0.0
-          and math.isclose(initializer.stddev, 1/math.sqrt(self.dim))):  # pytype: disable=module-attr (math.isclose not in PY2)
+          and math.isclose(initializer.stddev, 1/math.sqrt(self.dim))):
         initializer = None
 
-    return (
-        "TableConfig(vocabulary_size={vocabulary_size!r}, dim={dim!r}, "
-        "initializer={initializer!r}, optimizer={optimizer!r}, "
-        "combiner={combiner!r}, name={name!r})".format(
-            vocabulary_size=self.vocabulary_size,
-            dim=self.dim,
-            initializer=initializer,
-            optimizer=self.optimizer,
-            combiner=self.combiner,
-            name=self.name,)
-    )
+    return ("TableConfig(vocabulary_size={vocabulary_size!r}, dim={dim!r}, "
+            "initializer={initializer!r}, optimizer={optimizer!r}, "
+            "combiner={combiner!r}, name={name!r}, "
+            "quantization_config={quantization!r})".format(
+                vocabulary_size=self.vocabulary_size,
+                dim=self.dim,
+                initializer=initializer,
+                optimizer=self.optimizer,
+                combiner=self.combiner,
+                name=self.name,
+                quantization=self.quantization_config,
+            ))
+
+  def _set_table_descriptor(
+      self,
+      table_descriptor: tpu_embedding_configuration_pb2
+      .TPUEmbeddingConfiguration.TableDescriptor,
+      num_hosts: int,
+      learning_rate_index: Dict[Callable[[], Any], int]):
+    """Set the table descriptor from the table data."""
+    table_descriptor.name = self.name
+
+    # For small tables, we pad to the number of hosts so that at least one
+    # id will be assigned to each host.
+    table_descriptor.vocabulary_size = max(self.vocabulary_size, num_hosts)
+    table_descriptor.dimension = self.dim
+
+    parameters = table_descriptor.optimization_parameters
+
+    # We handle the learning rate separately here and don't allow the
+    # optimization class to handle this, as it doesn't know about dynamic
+    # rates.
+    if self.optimizer:
+      if callable(self.optimizer.learning_rate):
+        parameters.learning_rate.dynamic.tag = (
+            learning_rate_index[self.optimizer.learning_rate])
+      else:
+        parameters.learning_rate.constant = self.optimizer.learning_rate
+      if self.optimizer.low_dimensional_packing_status:
+        parameters.low_dimensional_packing_status = (
+            optimization_parameters_pb2.LowDimensionalPackingStatus.Status.ENABLED
+        )
+      # Use optimizer to handle the rest of the parameters.
+      self.optimizer._set_optimization_parameters(parameters)  # pylint: disable=protected-access
+
+    if self.quantization_config:
+      self.quantization_config._set_optimization_parameters(parameters)  # pylint: disable=protected-access
+
+
+@tf_export("tpu.experimental.embedding.RowIdInitializer")
+class RowIdInitializer:
+  """An initializer that initializes the table with vocabulary ids."""
+
+  def __init__(self, offset: int = 0):
+    self.offset = offset
+
+  def __call__(
+      self, shape: Union[Sequence[int], TensorShape], dtype: dtypes.DType
+  ) -> core.Tensor:
+    return math_ops.range(
+        start=self.offset, limit=self.offset + shape[0], delta=1, dtype=dtype
+    )[:, None] * array_ops.ones(shape, dtype=dtype)
 
 
 @tf_export("tpu.experimental.embedding.FeatureConfig")
-class FeatureConfig(object):
+class FeatureConfig:
   """Configuration data for one embedding feature.
 
   This class holds the configuration data for a single embedding feature. The
@@ -1033,7 +1419,8 @@ class FeatureConfig(object):
         has to match the shape (for ragged tensor, the input shape and output
         shape can mismatch). If not provided, the shape can be either provided
         to the `embedding.build` or auto detected at the runtime.
-      name: An optional name for the feature, useful for debugging.
+      name: An optional string used to name the table. Must be defined if
+        running on SparseCore.
 
     Returns:
       `FeatureConfig`.
@@ -1068,11 +1455,12 @@ class FeatureConfig(object):
   def __repr__(self):
     return ("FeatureConfig(table={table!r}, "
             "max_sequence_length={max_sequence_length!r}, "
-            "validate_weights_and_indices={"
-            "validate_weights_and_indices!r}, name={name!r})".format(
+            "validate_weights_and_indices={validate_weights_and_indices!r}, "
+            "output_shape={output_shape!r}, name={name!r})".format(
                 table=self.table,
                 max_sequence_length=self.max_sequence_length,
                 validate_weights_and_indices=self.validate_weights_and_indices,
+                output_shape=self.output_shape,
                 name=self.name))
 
 
@@ -1089,3 +1477,31 @@ def log_tpu_embedding_configuration(
   for line in str(config).splitlines():
     logging.info(line)
   logging.info("Done with log of TPUEmbeddingConfiguration.")
+
+
+def _sort_device_spec_strings(device_strings: Iterable[str]) -> List[str]:
+  sorted_specs = sorted(
+      (device_spec.DeviceSpecV2.from_string(spec) for spec in device_strings),
+      key=lambda s: (s.replica, s.task, s.device_index),
+  )
+  return [spec.to_string() for spec in sorted_specs]
+
+
+def get_list_of_hosts(strategy: tpu_strategy.TPUStrategy) -> List[Text]:
+  """Returns a sorted list of CPU devices for the remote jobs.
+
+  Args:
+    strategy: A TPUStrategy object.
+
+  Returns:
+    A sorted list of device host strings.
+  """
+
+  list_of_hosts = []
+  # Elsewehere we assume that the list of hosts is sorted.
+  for tpu_device in _sort_device_spec_strings(strategy.extended.worker_devices):
+    host = device_util.get_host_for_device(tpu_device)
+    if host not in list_of_hosts:
+      list_of_hosts.append(host)
+  assert len(list_of_hosts) == strategy.extended.num_hosts
+  return list_of_hosts
